/*===========================================================================



BOOST GAUGE
===========



Overview:
=========

CPU         :   Arduino Nano / ATmega328 / Old bootloader
Clock       :   XTAL 16MHz
Compiler    :   Arduino 1.8.10
Version     :   1.00
Date        :   2020-12-20
Author      :   Iacopo Giangrandi HB9DUL



Description:
============

This program implements a simple turbo boost gauge that reads the boost 
pressure from the OBD-II port of the car and displays it on a 12-LED ring. 
It's based on an Arduino Nano board. To interface to the OBD-II port, a 
commercial ELM327-USB adapter is used and it's modified by removing the 
RS232 to USB interface chip and hooking the Arduino directly in to the 
RS232 (TTL-level) RX and TX signals. Two jumpers were initially installed to 
separate the two boards for debugging, but their presence is no longer 
needed. The 5V power supply for the Arduino and the display is taken from 
the ELM327 board as well, but a small heat sink (thick copper wire) is added 
to the 78M05 voltage regulator to help it dealing with the higher load. 
Additional bypass capacitors are also added to the power supply to increase 
voltage stability, because when the engine is running a strange behavior of 
the board has been observed.

The firmware periodically polls the "intake manifold absolute pressure" 
(IMAP, service 0x00, PID 0x0B) and subtracts the "barometric pressure" (BP, 
service 0x00, PID 0x33) to get the boost pressure. Because the barometric 
pressure changes only with changes in elevation and atmospheric weather, 
it's polled much less frequently to save bandwidth. Standard OBD-II commands 
allow computing boost roughly between -101 (total vacuum) and +154 (1.54 bar 
of positive pressure), because both pressures (atmospheric and intake 
manifold) are coded with 8 bits from 0 to 255.

The Arduino internal LED is switched on during each IMAP reading so that the 
sample rate can be measured with an oscilloscope.

The boost pressure is displayed on a 12-LED ring of WS2812 (NeoPixel) RGB 
smart LEDs allowing a single wire communication. The top LED is always white 
and shows the 0 bar mark. Negative pressure is indicated in blue and spins 
counterclockwise, positive pressure is red and spins clockwise. Each LED 
represents 0.1 bars (configurable in KPA_PER_PIXEL), and fractions are 
displayed by varying the brightness if the last LED. If the positive 
pressure exceeds the number of pixels (one full turn), a second loop is 
displayed in yellow over the already present red one.

A photoresistor senses the ambient light and automatically changes the 
brightness of the display between two levels (BRIGHTNESS_LOW and 
BRIGHTNESS_HIGH) with linear interpolataion in-between.

An additional pushbutton allows resetting the trouble codes of the engine 
control unit (ECU) by using service 0x04 of the OBD-II standard. By pressing 
and holding the button, a light green ring will spin on the display as a 
progress bar. If the button is held down long enough to complete the ring, 
the ECU trouble codes are cleared and a complete green ring will be 
displayed until the button is released.

Communication errors with the OBD-II interface or with the car ECU are 
displayed by blinking purple LEDs on the ring. The number of LEDs 
corresponds to the error number. Each error blinks three times.



I/O connections:
================

NeoPixel ring display:
----------------------
D10    ----> Data               WS2812 data pin

ELM327 module:
--------------
D3     <---- RX_PIN             Data received from the ELM327 (ELM327_TX)
D2     ----> TX_PIN             Data sent to the ELM327 (ELM327_RX)

Button(s) and LEDs:
-------------------
D12    <---- SW_P1              Button, active low, requires internal pull-up
D13    ----> LED_PIN            Arduino default LED, active high, used only for debugging

Photoresistor (light sensor):
-----------------------------
D14/A0 <---- AMBIENT_LIGHT_IN   Photoresistor between Vdd and A0, 10k between A0 and GND.



============================================================================= */



// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <Adafruit_NeoPixel.h>
#include <SoftwareSerial.h>
#include "ELMduino.h"



// --------------------------------------------------------------------------
// Hardware related definitions, constants and global variables.
// --------------------------------------------------------------------------
#define         WELCOME_STR         "BOOST GAUGE - v1.00 - 2020-12-20 - (c) Iacopo Giangrandi"

#define         SW_P1               12                          // Pushbutton connected to digital pin 12, idle high, active low
#define         LED_PIN             13                          // Internal Arduino LED connected to digital pin 13, high=on

// General boost measurement
#define         DEFAULT_PRESSURE    101                         // Default atmospheric pressure is 101kPa
#define         INTERVAL_BP         500                         // Reads barometric pressure every 500 readings
#define         INTERVAL_BRT        10                          // Reads brightness every 10 readings

// NeoPixel ring display
#define         DISPLAY_PIN         10                          // NeoPixel ring connected on D10
#define         NUMPIXELS           12                          // Ring is 12 pixel long
#define         KPA_PER_PIXEL       10                          // Each pixel represents 10kPa
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(NUMPIXELS, DISPLAY_PIN, NEO_GRB + NEO_KHZ800); // NeoPixel object

// Light sensor and LED brightness related constants.
#define         AMBIENT_LIGHT_IN    14                          // Analog input for light readings
#define         LIGHT_MIN           750                         // Minimum ambient light (ADC value) below which the minimum brightness is used
#define         LIGHT_MAX           900                         // Maximum ambient light (ADC value) above which the maximum brightness is used
#define         BRIGHTNESS_MIN      16                          // Minimum display brighness (1..255)
#define         BRIGHTNESS_MAX      90                          // Maximum display brighness (1..255)
int brightness = BRIGHTNESS_MAX;                                // Brightness global variable

// ELM327 OBD-II interface
SoftwareSerial elm_port(3, 2);                                  // Software serial port: RX on D2, TX on D2
ELM327 myELM327;                                                // ELM327 object



// --------------------------------------------------------------------------
// This is the setup procedure executed once at startup. This procedure is
// automatically executed by Arduino.
// --------------------------------------------------------------------------
void setup()
{
    // Configure I/O lines.
    pinMode(SW_P1, INPUT_PULLUP);                               // Configure the input line for the pushbutton.
    pinMode(LED_PIN, OUTPUT);                                   // Digital pin 13 is an output for the LED.
    digitalWrite(LED_PIN, LOW);                                 // Turn the internal LED off.

    // Configure serial port (console).
    Serial.begin(115200);                                       // Set serial communication at 38400N81.
    Serial.println(F(WELCOME_STR));                             // Print welcome string.

    // Configure display
    pixels.begin();                                             // This initializes the NeoPixel library.
    set_brightness();
    display_welcome_sequence();

    // Configure ELM327
    elm_port.begin(38400);
    if (!myELM327.begin(elm_port))
    {
        Serial.println(F("ELM327 connection failed."));
        display_error(12);
        //while (1);
    }
}



// --------------------------------------------------------------------------
// This is the main program. It's executed automatically after the 
// initialization in an andless loop by Arduino.
// --------------------------------------------------------------------------
void loop()
{
    int imap = DEFAULT_PRESSURE;
    int bp = DEFAULT_PRESSURE;
    int boost;
    static int counter_bp = INTERVAL_BP;
    static int counter_brt = INTERVAL_BRT;

    counter_brt++;                                              // A brightness reading every 100 cycles is enough.
    if (counter_brt >= INTERVAL_BRT)
    {
        set_brightness();
        counter_brt = 0;
    }

    counter_bp++;                                               // A barometric pressure reading every 1000 cycles is enough.
    if (counter_bp >= INTERVAL_BP)
    {
        bp = elm327_read_bp();
        counter_bp = 0;
    }

    digitalWrite(LED_PIN, HIGH);                                // Turn the internal LED on during the IMAP reading to measure with an oscilloscope how often this is done.
    imap = elm327_read_imap();                                  // Reads absolute manifold pressure and calculate boost.
    boost = imap - bp;
    digitalWrite(LED_PIN, LOW);

    display_boost(boost, brightness);

    button_handler();                                           // Handle the button to clear the trouble codes of the ECU if the button is pressed long enough.

    delay(50);                                                  // Wait 50ms
}



// --------------------------------------------------------------------------
// This function will query the ELM327 and return the intake manifold 
// absolute pressure. This is a value in kPa between 0 and 255. If an error 
// occours, print_error() is executed and DEFAULT_PRESSURE is retourned to 
// avoid disply glitches.
// --------------------------------------------------------------------------
int elm327_read_imap(void)
{
    if (myELM327.queryPID(SERVICE_01, INTAKE_MANIFOLD_ABS_PRESSURE))    // Call service 0x01, PID 0x0B
    {
        int imap = myELM327.findResponse();                     // One byte expected, pressure in kPa

        if (myELM327.status == ELM_SUCCESS)
        {
            Serial.print(F("IMAP: "));
            Serial.println(imap);
            return(imap);
        }
    }
    elm_print_error();
    return(DEFAULT_PRESSURE);
}



// --------------------------------------------------------------------------
// This function will query the ELM327 and return the barometric pressure.
// This is a value in kPa between 0 and 255. If an error occours, 
// print_error() is executed and DEFAULT_PRESSURE is retourned to avoid 
// disply glitches.
// --------------------------------------------------------------------------
int elm327_read_bp(void)
{
    if (myELM327.queryPID(SERVICE_01, ABS_BAROMETRIC_PRESSURE)) // Call service 0x01, PID 0x33
    {
        int bp = myELM327.findResponse();                       // One byte expected, pressure in kPa

        if (myELM327.status == ELM_SUCCESS)
        {
            Serial.print(F("BP: "));
            Serial.println(bp);
            return(bp);
        }
    }
    elm_print_error();
    return(DEFAULT_PRESSURE);
}



// --------------------------------------------------------------------------
// This function will clear the ECU trouble codes through service 04 of the
// OBD-II interface.
// --------------------------------------------------------------------------
void elm327_clear_codes(void)
{
    char service_and_PID[] = "04";

    if (myELM327.queryPID(service_and_PID))                     // Call service 0x04, no response expected
    {
        if (myELM327.status == ELM_SUCCESS)
        {
            Serial.println(F("Codes cleared"));
            return;
        }
    }
    elm_print_error();
}



// --------------------------------------------------------------------------
// This procedure reads the ambient light via a photoresistor connected to 
// the A0 line. The photoresistor is connected between Vdd and A0 and a 10k 
// resistor to GND completes a voltage divider. 
// It sets the display brightness to a high (BRIGHTNESS_HIGH) or a low 
// (BRIGHTNESS_LOW) value when the ambient ligth goes above or below a 
// threshold a maximum or minimum value (LIGHT_MAX or LIGHT_MIN). Values 
// in-between are linearly interpolated.
// --------------------------------------------------------------------------
void set_brightness(void)
{
    long int adc_value;

    adc_value = analogRead(AMBIENT_LIGHT_IN);                   // Read the voltage on LIGHT_IN.

    if (adc_value < LIGHT_MIN)
        brightness = BRIGHTNESS_MIN;
    else if (adc_value > LIGHT_MAX)
        brightness = BRIGHTNESS_MAX;
    else
        brightness = ((adc_value - LIGHT_MIN) * (BRIGHTNESS_MAX - BRIGHTNESS_MIN) / (LIGHT_MAX - LIGHT_MIN)) + BRIGHTNESS_MIN;

    Serial.print(F("BRT: "));                                   // Print value on the console
    Serial.print(adc_value);
    Serial.print(F(", "));
    Serial.println(brightness);
}



// --------------------------------------------------------------------------
// This procedure displays the boost pressure on the neopixel ring. 
// boost_pressure is usually between -101 (total vacuum) and +154 (1.54 bar 
// of positive pressure). Brightness defines the brightness of the ring and 
// must be between 0 and 255. 
// Negative pressure is blue, positive is red. For positive pressure (only) 
// exceding the maximum amount of pixels (defined in NUMPIXELS), a second 
// loop is displayed in yellow.
// --------------------------------------------------------------------------
void display_boost(int boost_pressure, int brightness)
{
    int i;
  
    for(i = 0; i < NUMPIXELS; i++)
        pixels.setPixelColor(i, pixels.Color(0, 0, 0));         // All pixels off.
  
    if (boost_pressure >= 0)
    {
        pixels.setPixelColor(0, pixels.Color(brightness/2, brightness/2, brightness/2));  // Zero is marked white.
        
        i = 1;                                                  // Start from second (i=1) pixel as zero is the white mark.
        while (boost_pressure > KPA_PER_PIXEL)                  // First process all full value pixels.
        {
            pixels.setPixelColor(i, pixels.Color(brightness, 0, 0));    // Positive pressure is red.
            boost_pressure -= KPA_PER_PIXEL;                    // Decrease boost by one full pixel pressure.
            i++;                                                // Select next pixel.
            if (i >= (NUMPIXELS - 1))                           // Keep going until the maximum number of pixels is reached.
                break;
        }
        if (boost_pressure <= KPA_PER_PIXEL)                    // If prssure is low enough, we can display it with the pixel we have.
            pixels.setPixelColor(i, pixels.Color((brightness * boost_pressure / KPA_PER_PIXEL), 0, 0)); // Display last pixel with proportional value and we are done.
        else
        {                                                       // We need more pixels and we will loop again.
            pixels.setPixelColor(i, pixels.Color(brightness, 0, 0));    // Display last full pixel of the circle in red.
            boost_pressure -= KPA_PER_PIXEL;                    // Decrease boost by one full pixel pressure.
            i = 0;                                              // And start again for a second loop from the first (i=0) pixel overriding the white mark.
            while (boost_pressure > KPA_PER_PIXEL)              // First process all full value pixels.
            {
                pixels.setPixelColor(i, pixels.Color(brightness, brightness, 0));   // Second loop is displayed as yellow.
                boost_pressure -= KPA_PER_PIXEL;                // Decrease boost by one full pixel pressure.
                i++;                                            // Select next pixel.
                if (i >= (NUMPIXELS - 1))                       // Make sure we are not exceeding the maximum number of pixels.
                {
                    if (boost_pressure > KPA_PER_PIXEL)         // And make sure the remaining pressure doesn't exceed the maximum pixel value.
                        boost_pressure = KPA_PER_PIXEL;
                    break;                                      // Avoid processing more pixels.
                }
            }     
            pixels.setPixelColor(i, pixels.Color((brightness * boost_pressure / KPA_PER_PIXEL), (brightness * boost_pressure / KPA_PER_PIXEL), 0)); // Display last pixel with proportional value.
        }
    }
    
    else
    
    {
        boost_pressure = -boost_pressure;                       // Boost is negative: take absolute value.
        pixels.setPixelColor(0, pixels.Color(brightness/2, brightness/2, brightness/2));  // Zero is marked white.
        
        i = NUMPIXELS - 1;                                      // Start from last pixel.
        while (boost_pressure > KPA_PER_PIXEL)                  // First process all full value pixels.
        {
            pixels.setPixelColor(i, pixels.Color(0, 0, brightness));    // Negative pressure is blue.
            boost_pressure -= KPA_PER_PIXEL;                    // Decrease boost by one full pixel pressure.
            i--;                                                // Select previous pixel.
            if (i == 0)                                         // Make sure there are still pixels to process.
            {
                if (boost_pressure > KPA_PER_PIXEL)             // And make sure the remaining pressure doesn't exceed the maximum pixel value.
                    boost_pressure = KPA_PER_PIXEL;
                break;                                          // Avoid processing more pixels (even if normally pressure cannot be less than zero).
            }
        }
        pixels.setPixelColor(i, pixels.Color(0, 0, (brightness * boost_pressure / KPA_PER_PIXEL))); // Display last pixel with proportional value.
    }
  
    pixels.show();                                              // This sends the updated pixel color to the hardware.
}



// --------------------------------------------------------------------------
// This procedure will print the ELM327 error on the console and then blink 
// the corresponding error number on the display through display_error().
// --------------------------------------------------------------------------
void elm_print_error()
{
    Serial.print("Received: ");                                 // Print the data that was received.
    for (byte i = 0; i < myELM327.recBytes; i++)
        Serial.write(myELM327.payload[i]);
    Serial.println();

    Serial.print(F("ELM Error "));                              // Print the error number... 
    Serial.print(myELM327.status);

    Serial.print(F(": "));                                      // ...and its description
    if (myELM327.status == ELM_SUCCESS)
        Serial.println(F("Success"));
    else if (myELM327.status == ELM_NO_RESPONSE)
        Serial.println(F("No response"));
    else if (myELM327.status == ELM_BUFFER_OVERFLOW)
        Serial.println(F("Buffer overflow"));
    else if (myELM327.status == ELM_UNABLE_TO_CONNECT)
        Serial.println(F("Unable to connect"));
    else if (myELM327.status == ELM_NO_DATA)
        Serial.println(F("No data"));
    else if (myELM327.status == ELM_STOPPED)
        Serial.println(F("Stopped"));
    else if (myELM327.status == ELM_TIMEOUT)
        Serial.println(F("Timeout"));
    else if (myELM327.status == ELM_TIMEOUT)
        Serial.println(F("General error"));

    display_error(myELM327.status);                             // Blink the error number on the display
}



// --------------------------------------------------------------------------
// This procedure will make the number of LED specified in err_num blink for 
// three times. This is to display an error number. Color is purple.
// Negative or large values are displyed with all LEDs blinking.
// --------------------------------------------------------------------------
void display_error(int err_num)
{
    int i, j;

    if (err_num == 0)                                           // 0 means "no error", return.
        return;

    if ((err_num < 0) || (err_num > NUMPIXELS))                 // We only have 12 pixels: only errors from 1 to 12 can be displayed.
        err_num = NUMPIXELS;

    for (j = 0; j < 3; j++)                                     // Blink a few times.
    {
        for(i = 0; i < NUMPIXELS; i++)                          // Set corresponding number of pixels to purple.
        {
            if (i < err_num)
                pixels.setPixelColor(i, pixels.Color(brightness, 0, brightness)); 
            else
                pixels.setPixelColor(i, pixels.Color(0, 0, 0));         
        }
        pixels.show();
        delay(100);
        for(i = 0; i < NUMPIXELS; i++)                          // Clear the display.
            pixels.setPixelColor(i, pixels.Color(0, 0, 0));
        pixels.show();
        delay(333);
    }
    delay(1000);
}



// --------------------------------------------------------------------------
// This procedure checks if the button is pressed. If it is, it will spin a 
// light green ring as a progress bar as long as the button is held down. 
// If the button is held down long enough to complete the ring, the ECU 
// trouble codes will be cleared and a complete green ring will be displayed 
// until the button is released (or an error code will blink instead).
// --------------------------------------------------------------------------
void button_handler(void)
{
    int i, j;

    if (digitalRead(SW_P1) == HIGH)                             // If the button is not pressed there is nothing to do.
        return;

    j = 0;                                                      // Spin a light green ring while the button is held down, like a progress bar.
    while (digitalRead(SW_P1) == LOW)
    {
        j++;
        for(i = 0; i < NUMPIXELS; i++)
        {
            if (i < j)
                pixels.setPixelColor(i, pixels.Color(0, brightness, brightness)); 
            else
                pixels.setPixelColor(i, pixels.Color(0, 0, 0));         
        }
        pixels.show();
        delay(250);

        if (j >= NUMPIXELS)                                     // Break when the ring is complete, even if the button is still down.
            break;
    }

    if (j >= NUMPIXELS)                                         // If we reached the complete ring (button held down lon enough) we can reset our trouble codes and display a fully green ring.
    {
        elm327_clear_codes();                                   // Clear the ECU trouble codes.

        if (myELM327.status == ELM_SUCCESS)                     // Display the green ring only if the operation is successful (if not, the error code will be displayed by elm327_clear_codes())
        {
            for(i = 0; i < NUMPIXELS; i++)
                pixels.setPixelColor(i, pixels.Color(0, brightness, 0));
            pixels.show();                                      // This sends the updated pixel color to the hardware.
        }

        while (digitalRead(SW_P1) == LOW) ;                     // Wait here until the button is released.
    }

    for(i = 0; i < NUMPIXELS; i++)                              // Clear the display.
        pixels.setPixelColor(i, pixels.Color(0, 0, 0));
    pixels.show();

    delay(100);                                                 // Final delay to debounce the button.
}



// --------------------------------------------------------------------------
// Display a spinning red pluse that gently lights up the white zero mark 
// of the regular boost display.
// --------------------------------------------------------------------------
void display_welcome_sequence(void)
{
    int i, j, k;

    byte wave[] = { byte(brightness / 4),                       // A wave pulse composed by progressively brighter and dimmer LEDs. This array cannot be empty (division by zero).
                    byte(brightness / 2), 
                    byte(brightness), 
                    byte(brightness), 
                    byte(brightness / 2), 
                    byte(brightness / 4) };

    for (j = 0; j < NUMPIXELS + sizeof(wave); j++)              // Shift the wave through the whole ring untill it's completely out.
    {
        for(i = 0; i < NUMPIXELS; i++)                          // For every iteration, compute all 12 pixels
        {
            k = i - j + sizeof(wave) - 1;                       // Calulate the shift k, where the wave will land on the pixel array
            if ((k >= 0) && (k < sizeof(wave)))
            {
                pixels.setPixelColor(i, pixels.Color(wave[k], 0, 0));
            }
            else
            {
                pixels.setPixelColor(i, pixels.Color(0, 0, 0));
            }
        }

        if (j >= NUMPIXELS)                                     // When the wave arrives at theend of the array, gradually switch on the white 0-bar mark.
        {
            k = ((j - NUMPIXELS + 1) * brightness) / sizeof(wave);
            k /= 2;                                             // White is too bright with all three RGB LEDs lit at the same time.
            pixels.setPixelColor(0, pixels.Color(k, k, k));
        }

        pixels.show();
        delay(50);
    }
}
